import { pan, zoom } from 'pan-zoom';

export default class PanZoom {
  #storageKey = 'pan-zoom';
  #zoomFactor = 0.25;

  constructor(svg) {
    this.#restorePanZoomVal(svg);
    this.#addEventListeners(svg);
    this.#observePanZoomChanges(svg);
  }

  #storePanZoomVal(transformMatrix) {
    localStorage.setItem(this.#storageKey, transformMatrix);
  }

  #observePanZoomChanges(svg) {
    const observer =
      new MutationObserver(() => this.#storePanZoomVal(svg.style.transform));

    observer.observe(svg, { attributeFilter: ['style'] });
  }

  #restorePanZoomVal(svg) {
    const storedPanZoomVal = localStorage.getItem(this.#storageKey);

    if (storedPanZoomVal) {
      svg.style.transform = storedPanZoomVal;
    }
  }

  #addEventListeners(svg) {
    svg.addEventListener('wheel', e => zoom(svg, e, this.#zoomFactor), { passive: false });
    svg.addEventListener('pointerdown', e => pan(svg, e), { passive: false });
  }
};
